//
// Simple shadow maps in OpenGL
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "libTexture.h"
#include	"Vector3D.h"
#include	"Vector4D.h"
#include	"GlslProgram.h"

Vector3D    eye    ( 7, 5, 8 );       	// camera position
Vector3D    light  ( 5, 0, 4 );      	// light position
Vector3D	center ( 0, 0, 0 );			// lookAt position
Vector3D	rot    ( 0, 0, 0 );
int			mouseOldX     = 0;
int			mouseOldY     = 0;
float       angle         = 0;
int			shadowMapSize = 512;
GLuint    	ambMap;                     // ambient texture
GLuint    	diffuseMap;                 // diffuse texture
GLuint		shadowMap;
int			width, height;

GlslProgram	sh;

float   mv [16];                        // to hold modelview matrix used when rendering from light
float   pr [16];                        // to hold projection matrix used when rendering from light

/////////////////////////// Some useful shadowing operations ////////////////////
/////////////////////////////////////////////////////////////////////////////////

void	renderScene ();
void 	reshape     ( int w, int h );

void createShadowMap ()
{
								// create depth texture
	glGenTextures   ( 1, &shadowMap );
	glBindTexture   ( GL_TEXTURE_2D, shadowMap );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

	glTexParameteri  ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB );
	glTexParameteri  ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL );

	glTexGeni        ( GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR );
	glTexGeni        ( GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR );
	glTexGeni        ( GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR );
	glTexGeni        ( GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR );

	glTexImage2D     ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapSize, shadowMapSize, 0,
                       GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL );
}

void init ()
{
    glClearColor   ( 0.0, 0.0, 0.0, 1.0 );
    glClearStencil ( 0 );
    glEnable       ( GL_DEPTH_TEST );
    glDepthFunc    ( GL_LESS );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void	renderToShadowMap ()
{
	glDisable ( GL_TEXTURE_2D );

	glEnable        ( GL_POLYGON_OFFSET_FILL );
	glPolygonOffset ( 4, 4 );

									// setup projection
	glViewport ( 0, 0, shadowMapSize, shadowMapSize );
	glClear    ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();

    gluPerspective ( 120.0, 1, 0.1, 60.0 );
   	gluLookAt      ( light.x, light.y, light.z,		// eye
					 center.x, center.x, center.z,	// center
					 0, 0, 1 );						// up

	glMatrixMode   ( GL_MODELVIEW );
	glLoadIdentity ();

                                        // get modelview and projections matrices
    glGetFloatv ( GL_MODELVIEW_MATRIX,  mv );
    glGetFloatv ( GL_PROJECTION_MATRIX, pr );

										// now render scene from light position
	renderScene ();
										// restore state
	glBindTexture    ( GL_TEXTURE_2D, shadowMap );
	glCopyTexImage2D ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, shadowMapSize, shadowMapSize, 0 );
										
	glDisable        ( GL_POLYGON_OFFSET_FILL );
	glEnable         ( GL_TEXTURE_2D );
}

void	drawQuad ( unsigned tex )
{
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glColor4f     ( 1, 1, 1, 1 );
    glBindTexture ( GL_TEXTURE_2D, tex );
	glEnable      ( GL_TEXTURE_2D );
    glBegin       ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1,   0 );
        glVertex2f   ( shadowMapSize, 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( shadowMapSize, shadowMapSize );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, shadowMapSize );

    glEnd   ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, shadowMapSize, 0, shadowMapSize, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix

	glDisable   ( GL_DEPTH_TEST );
	glDepthMask ( GL_FALSE );
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix

	glEnable    ( GL_DEPTH_TEST );
	glDepthMask ( GL_TRUE );
}

void	renderScene ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glMatrixMode       ( GL_MODELVIEW );
    glColor3f          ( 0.5, 0.5, 0.5 );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

	glMatrixMode    ( GL_MODELVIEW );
	glPushMatrix    ();
	glRotatef       ( 90, 1, 0, 0 );
	glTranslatef    ( 0, 0.5, 0 );
	glutSolidTeapot ( 2.5 );
	glPopMatrix     ();
                                                // draw unlit floor
    glBindTexture      ( GL_TEXTURE_2D, ambMap );

    glBegin ( GL_QUADS );

    glTexCoord2f ( 0, 0 );
    glVertex3f   ( -8, -8, -1 );

    glTexCoord2f ( 1, 0 );
    glVertex3f   ( 8, -8, -1 );

    glTexCoord2f ( 1, 1 );
    glVertex3f   ( 8, 8, -1 );

    glTexCoord2f ( 0, 1 );
    glVertex3f   ( -8, 8, -1 );

    glEnd   ();
}

void display ()
{
	renderToShadowMap ();						// compute shadow map
												// clear buffers
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	reshape ( width, height );					// setup modelview and projection
	
	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

												// setup shadowing
	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, shadowMap );
	glTexParameteri    ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri    ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexEnvi          ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glActiveTextureARB ( GL_TEXTURE1_ARB );
	
	glColor4f          ( 1, 1, 1, 1 );
										// set correct texcoord transform
    glMatrixMode  ( GL_TEXTURE );

    glLoadIdentity ();
    glTranslatef   ( 0.5, 0.5, 0.5 );     // remap from [-1,1]^2 to [0,1]^2
    glScalef       ( 0.5, 0.5, 0.5 );
    glMultMatrixf  ( pr );
    glMultMatrixf  ( mv );

	glActiveTextureARB ( GL_TEXTURE0_ARB );

	sh.bind    ();
	renderScene ();
	sh.unbind  ();

	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glDisable          ( GL_TEXTURE_2D   );
	glActiveTextureARB ( GL_TEXTURE0_ARB );

                                                // draw the light
    glMatrixMode ( GL_MODELVIEW );
    glPopMatrix  ();
    glPushMatrix ();

    glTranslatef       ( light.x, light.y, light.z );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glDisable          ( GL_TEXTURE_2D );
    glutSolidSphere    ( 0.1f, 15, 15 );
    glPopMatrix        ();

    glMatrixMode ( GL_TEXTURE );
    glPopMatrix  ();

    glMatrixMode ( GL_MODELVIEW );

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
    glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();

    gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 0.1, 60.0 );
    gluLookAt      ( eye.x, eye.y, eye.z,    		// eye
                     center.x, center.y, center.z,  // center
                     0, 0, 1 );              		// up

	glMatrixMode   ( GL_MODELVIEW );
    glLoadIdentity ();

    width  = w;
    height = h;
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
}

void    animate ()
{
    angle  = 0.002f * glutGet ( GLUT_ELAPSED_TIME );

    light.x = 3*cos ( angle );
    light.y = 3*sin ( angle );
    light.z = 5 + 0.9 * sin ( angle / 3 );

    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL );
    glutInitWindowSize  ( 700, 600 );


                                // create window
    glutCreateWindow ( "OpenGL PCF Shadow Map demo" );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
    glutIdleFunc     ( animate );

    init           ();
    initExtensions ();

	assertExtensionsSupported ( "GL_ARB_shadow GL_ARB_depth_texture" );

	createShadowMap ();

    ambMap      = createTexture2D ( true,  "../../Textures/oak.bmp" );
    diffuseMap  = createTexture2D ( true,  "../../Textures/Oxidated.jpg" );

    if ( !sh.loadShaders ( "sh-2.vsh", "sh-2.fsh" ) )
    {
        printf ( "Error loading sh-2 shaders:\n%s\n", sh.getLog ().c_str () );

        return 1;
    }
	
	sh.bind       ();
	sh.setTexture ( "mainMap",   0 );
	sh.setTexture ( "shadowMap", 1 );
	sh.unbind     ();
	
	glBindTexture ( GL_TEXTURE_2D, 0 );
	
    glutMainLoop ();

    return 0;
}
